import os.path
import tempfile
import unittest

from symbolchain.core.CryptoTypes import PrivateKey
from symbolchain.core.PrivateKeyStorage import PrivateKeyStorage


class PrivateKeyStorageTest(unittest.TestCase):
    # region save

    def _assert_can_save_private_keys(self, private_key_names, password):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage = PrivateKeyStorage(temp_directory, password)

            # Act:
            for name in private_key_names:
                storage.save(name, PrivateKey.random())

            # Assert:
            self.assertEqual(len(private_key_names), len(os.listdir(temp_directory)))
            for name in private_key_names:
                self.assertTrue(os.path.exists(os.path.join(temp_directory, '{}.pem'.format(name))))

    def test_can_save_single_private_key_without_encryption(self):
        self._assert_can_save_private_keys(['alpha'], None)

    def test_can_save_single_private_key_with_encryption(self):
        self._assert_can_save_private_keys(['alpha'], 'password')

    def test_can_save_multiple_private_keys_without_encryption(self):
        self._assert_can_save_private_keys(['alpha', 'beta', 'gamma'], None)

    def test_can_save_multiple_private_keys_with_encryption(self):
        self._assert_can_save_private_keys(['alpha', 'beta', 'gamma'], 'password')

    # endregion

    # region load

    def _assert_can_roundtrip_private_keys(self, private_key_names, password):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage = PrivateKeyStorage(temp_directory, password)

            private_keys = [PrivateKey.random() for _ in private_key_names]

            # Act:
            for i, private_key in enumerate(private_keys):
                storage.save(private_key_names[i], private_key)

            loaded_private_keys = [storage.load(name) for name in private_key_names]

            # Assert:
            self.assertEqual(len(private_key_names), len(os.listdir(temp_directory)))
            for i, private_key in enumerate(private_keys):
                self.assertTrue(private_key, loaded_private_keys[i])

    def test_can_roundtrip_single_private_key_without_encryption(self):
        self._assert_can_roundtrip_private_keys(['alpha'], None)

    def test_can_roundtrip_single_private_key_with_encryption(self):
        self._assert_can_roundtrip_private_keys(['alpha'], 'password')

    def test_can_roundtrip_multiple_private_keys_without_encryption(self):
        self._assert_can_roundtrip_private_keys(['alpha', 'beta', 'gamma'], None)

    def test_can_roundtrip_multiple_private_keys_with_encryption(self):
        self._assert_can_roundtrip_private_keys(['alpha', 'beta', 'gamma'], 'password')

    # endregion

    # region failure cases

    def test_cannot_load_private_key_that_does_not_exist(self):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage = PrivateKeyStorage(temp_directory)
            storage.save('foo', PrivateKey.random())

            # Sanity:
            self.assertEqual(1, len(os.listdir(temp_directory)))

            # Act + Assert:
            with self.assertRaises(FileNotFoundError):
                storage.load('bar')

    def _assert_cannot_load_private_key_with_wrong_password(self, password1, password2, expected_error):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage1 = PrivateKeyStorage(temp_directory, password1)
            storage1.save('foo', PrivateKey.random())

            storage2 = PrivateKeyStorage(temp_directory, password2)

            # Act + Assert:
            with self.assertRaises(expected_error):
                storage2.load('foo')

    def test_cannot_load_unencrypted_private_key_with_password(self):
        self._assert_cannot_load_private_key_with_wrong_password(None, 'password', TypeError)

    def test_cannot_load_encrypted_private_key_without_password(self):
        self._assert_cannot_load_private_key_with_wrong_password('password', None, TypeError)

    def test_cannot_load_encrypted_private_key_with_wrong_password(self):
        self._assert_cannot_load_private_key_with_wrong_password('password', 'password2', ValueError)

    # endregion



import tempfile
import unittest

from symbolchain.core.CryptoTypes import Hash256, Signature
from symbolchain.core.QrSignatureStorage import QrSignatureStorage
from symbolchain.tests.test.NemTestUtils import NemTestUtils


def write_random_qrcode(directory, name, data_size):
    # abuse transaction_hash by changing its underlying bytes
    transaction_hash = NemTestUtils.randcryptotype(Hash256)
    transaction_hash.bytes = NemTestUtils.randbytes(data_size)
    QrSignatureStorage(directory).save(name, transaction_hash, [])


class QrSignatureStorageTest(unittest.TestCase):
    def _assert_can_roundtrip_signatures(self, num_signatures):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            storage = QrSignatureStorage(temp_directory)

            transaction_hash = NemTestUtils.randcryptotype(Hash256)
            signatures = [NemTestUtils.randcryptotype(Signature) for _ in range(0, num_signatures)]
            storage.save('foo', transaction_hash, signatures)

            # Act:
            (loaded_transaction_hash, loaded_signatures) = storage.load('foo')

            # Assert:
            self.assertEqual(transaction_hash, loaded_transaction_hash)
            self.assertEqual(signatures, loaded_signatures)

    def test_can_roundtrip_zero_signatures(self):
        self._assert_can_roundtrip_signatures(0)

    def test_can_roundtrip_single_signature(self):
        self._assert_can_roundtrip_signatures(1)

    def test_can_roundtrip_multiple_signatures(self):
        self._assert_can_roundtrip_signatures(5)

    def _assert_cannot_load_qrcode(self, data_size):
        # Arrange:
        with tempfile.TemporaryDirectory() as temp_directory:
            write_random_qrcode(temp_directory, 'foo', data_size)
            storage = QrSignatureStorage(temp_directory)

            with self.assertRaises(ValueError):
                storage.load('foo')

    def test_cannot_load_qrcode_containing_insufficient_data(self):
        self._assert_cannot_load_qrcode(Hash256.SIZE - 1)  # less than transaction_hash

    def test_cannot_load_qrcode_containing_partial_signature(self):
        self._assert_cannot_load_qrcode(Hash256.SIZE + Signature.SIZE * 2 + Signature.SIZE // 2)  # 2.5 signatures



from binascii import unhexlify


class BufferWriter:
    """Writes data to an in memory buffer."""

    def __init__(self, byte_order='little'):
        """Creates a writer with specified byte order."""
        self.byte_order = byte_order
        self.buffer = bytes()

    def write_int(self, value, count):
        """Writes an integer."""
        self.buffer += value.to_bytes(count, self.byte_order)

    def write_string(self, value):
        """Writes a string."""
        self.buffer += value.encode('utf8')

    def write_hex_string(self, value):
        """Writes a hex string."""
        self.buffer += unhexlify(value)

    def write_bytes(self, value):
        """Writes bytes."""
        self.buffer += value



from binascii import hexlify, unhexlify


class ByteArray:
    """Represents a fixed size byte array."""

    def __init__(self, fixed_size, array_input, tag=None):
        """Creates a byte array from bytes or hex string."""
        raw_bytes = array_input
        if isinstance(raw_bytes, str):
            raw_bytes = unhexlify(raw_bytes)

        if fixed_size != len(raw_bytes):
            raise ValueError('bytes was size {} but must be {}'.format(len(raw_bytes), fixed_size))

        self.bytes = raw_bytes
        self.__tag = tag

    def __eq__(self, other):
        # pylint: disable=protected-access
        return isinstance(other, ByteArray) and self.bytes == other.bytes and self.__tag == other.__tag

    def __str__(self):
        return hexlify(self.bytes).decode('utf8').upper()



import base64

import sha3

from ..ByteArray import ByteArray
from ..CryptoTypes import Hash256
from ..Network import Network as BasicNetwork


class Address(ByteArray):
    """Represents a symbol address."""

    SIZE = 24

    def __init__(self, address):
        """Creates an address from a decoded or encoded address."""
        raw_bytes = address
        if isinstance(address, str):
            raw_bytes = base64.b32decode(address + 'A')[0:-1]
        elif isinstance(address, Address):
            raw_bytes = address.bytes

        super().__init__(self.SIZE, raw_bytes, Address)

    def __str__(self):
        return base64.b32encode(self.bytes + bytes(0)).decode('utf8')[0:-1]


class Network(BasicNetwork):
    """Represents a symbol network."""

    def __init__(self, name, identifier, generation_hash_seed=None):
        """Creates a new network with the specified name, identifier byte and generation hash seed."""
        super().__init__(name, identifier)
        self.generation_hash_seed = generation_hash_seed

    def address_hasher(self):
        return sha3.sha3_256()

    def create_address(self, address_without_checksum, checksum):
        return Address(address_without_checksum + checksum[0:3])


Network.PUBLIC = Network('public', 0x68, Hash256('57F7DA205008026C776CB6AED843393F04CD458E0AA2D9F1D5F31A402072B2D6'))
Network.PRIVATE = Network('private', 0x78)
Network.PUBLIC_TEST = Network('public_test', 0x98, Hash256('45FBCF2F0EA36EFA7923C9BC923D6503169651F7FA4EFC46A8EAF5AE09057EBD'))
Network.PRIVATE_TEST = Network('private_test', 0xA8)
Network.NETWORKS = [Network.PUBLIC, Network.PRIVATE, Network.PUBLIC_TEST, Network.PRIVATE_TEST]



from ..CryptoTypes import PublicKey
from .Transaction import Transaction


class ImportanceTransferTransaction(Transaction):
    """Represents an importance transfer transaction."""
    NAME = 'importance-transfer'
    TYPE = 0x0801

    TYPE_HINTS = {
        'signer_public_key': PublicKey,
        'remote_account_public_key': PublicKey
    }

    def __init__(self, network):
        """Creates an importance transfer transaction for the specified network."""
        super().__init__(network, ImportanceTransferTransaction.TYPE)

        self.mode = 0
        self.remote_account_public_key = None

    @property
    def fee(self):
        """Gets the (minimum) fee."""
        return 150000

    def serialize_custom(self, writer):
        writer.write_int(self.mode, 4)

        writer.write_int(PublicKey.SIZE, 4)
        writer.write_bytes(self.remote_account_public_key.bytes)

    @staticmethod
    def field_names():
        return ['mode', 'remote_account_public_key']



import hashlib
import hmac

from mnemonic import Mnemonic

from .BufferWriter import BufferWriter
from .CryptoTypes import PrivateKey


class Bip32Node():
    """Representation of a BIP32 node."""

    def __init__(self, key, data):
        """Creates a BIP32 node around a key and data."""
        hmac_result = hmac.new(key, data, hashlib.sha512).digest()

        self.private_key = PrivateKey(hmac_result[0:PrivateKey.SIZE])
        self.chain_code = hmac_result[PrivateKey.SIZE:]

    def derive_one(self, identifier):
        """Derives a direct child node with specified identifier."""
        hmac_data_writer = BufferWriter('big')
        hmac_data_writer.write_int(0, 1)
        hmac_data_writer.write_bytes(self.private_key.bytes)
        hmac_data_writer.write_int(0x80000000 | identifier, 4)
        return Bip32Node(self.chain_code, hmac_data_writer.buffer)

    def derive_path(self, path):
        """Derives a descendent node with specified path."""
        next_node = self
        for identifier in path:
            next_node = next_node.derive_one(identifier)

        return next_node


class Bip32():
    """Factory of BIP32 root nodes """

    def __init__(self, curve_name='ed25519', mnemonic_language='english'):
        """Creates a BIP32 root node factory."""
        self.root_hmac_key = (curve_name + ' seed').encode('utf8')
        self.mnemonic_language = mnemonic_language

    def from_seed(self, seed):
        """Creates a BIP32 root node from a seed."""
        return Bip32Node(self.root_hmac_key, seed)

    def from_mnemonic(self, mnemonic, password):
        """Creates a BIP32 root node from a BIP39 mnemonic and password."""
        return self.from_seed(Mnemonic(self.mnemonic_language).to_seed(mnemonic, password))



import sha3

from ..CryptoTypes import Hash256


class MerkleHashBuilder:
    """Builder for creating a merkle hash."""

    def __init__(self):
        """Creates a merkle hash builder."""
        self.hashes = []

    def update(self, component_hash):
        """Adds a hash to the merkle hash."""
        self.hashes.append(component_hash.bytes)

    def final(self):
        """Calculates the merkle hash."""
        if not self.hashes:
            return Hash256.zero()

        num_remaining_hashes = len(self.hashes)
        while num_remaining_hashes > 1:
            i = 0
            while i < num_remaining_hashes:
                hasher = sha3.sha3_256()
                hasher.update(self.hashes[i])

                if i + 1 < num_remaining_hashes:
                    hasher.update(self.hashes[i + 1])
                else:
                    # if there is an odd number of hashes, duplicate the last one
                    hasher.update(self.hashes[i])
                    num_remaining_hashes += 1

                self.hashes[i // 2] = hasher.digest()
                i += 2

            num_remaining_hashes //= 2

        return Hash256(self.hashes[0])



import unittest
from functools import reduce

from symbolchain.core.CryptoTypes import PublicKey
from symbolchain.core.nis1.Network import Address, Network
from symbolchain.core.nis1.TransferTransaction import TransferTransaction
from symbolchain.tests.test.BasicNisTransactionTest import BasicNisTransactionTest, NisTransactionTestDescriptor

FOO_NETWORK = Network('foo', 0x54)


class TransferTransactionTest(BasicNisTransactionTest, unittest.TestCase):
    # region basic

    @staticmethod
    def get_test_descriptor():
        return NisTransactionTestDescriptor(TransferTransaction, 'transfer', 0x0101, FOO_NETWORK)

    def test_can_create(self):
        # Act:
        transaction = TransferTransaction(FOO_NETWORK)

        # Assert:
        self.assertEqual(0x0101, transaction.type)
        self.assertEqual(0x54000001, transaction.version)
        self.assertEqual(0, transaction.timestamp)
        self.assertEqual(None, transaction.signer_public_key)

        self.assertEqual(None, transaction.recipient_address)
        self.assertEqual(0, transaction.amount)
        self.assertEqual(None, transaction.message)

        # - properties
        self.assertEqual(50000, transaction.fee)
        self.assertEqual(60 * 60, transaction.deadline)

    # endregion

    # region properties

    def test_fee_is_updated_with_amount(self):
        # Arrange:
        transaction = TransferTransaction(FOO_NETWORK)

        # Act:
        transaction.amount = 15 * 10000000000 - 1

        # Assert:
        self.assertEqual(700000, transaction.fee)

    def test_fee_is_updated_with_message(self):
        # Arrange:
        transaction = TransferTransaction(FOO_NETWORK)

        # Act:
        transaction.message = [0, 1, 2] * 32 + [3] * 31

        # Assert:
        self.assertEqual(250000, transaction.fee)

    def test_fee_is_updated_with_amount_and_message(self):
        # Arrange:
        transaction = TransferTransaction(FOO_NETWORK)

        # Act:
        transaction.amount = 15 * 10000000000 - 1
        transaction.message = [0, 1, 2] * 32 + [3] * 31

        # Assert:
        self.assertEqual(900000, transaction.fee)

    def test_can_set_message_as_bytes(self):
        # Arrange:
        transaction = TransferTransaction(FOO_NETWORK)

        # Act:
        transaction.message = b'Hello!'

        # Assert:
        self.assertEqual(b'Hello!', transaction.message)

    def test_can_set_message_as_string(self):
        # Arrange:
        transaction = TransferTransaction(FOO_NETWORK)

        # Act:
        transaction.message = 'Hello!'

        # Assert:
        self.assertEqual(b'Hello!', transaction.message)

    # endregion

    # region serialize

    @staticmethod
    def _create_transfer_for_serialization_tests(include_message=False):
        transaction = TransferTransaction(FOO_NETWORK)
        transaction.timestamp = 12345
        transaction.signer_public_key = PublicKey('D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0')

        transaction.recipient_address = Address('TCFGSLITSWMRROU2GO7FPMIUUDELUPSZUNUEZF33')
        transaction.amount = 15 * 10000000000 - 1

        if include_message:
            transaction.message = [0x44, 0x55, 0x98, 0x12, 0x71, 0xAB, 0x72]

        return transaction

    def test_can_serialize_without_message(self):
        # Arrange:
        transaction = self._create_transfer_for_serialization_tests(False)

        # Act:
        buffer = transaction.serialize()

        # Assert:
        expected_buffers = [
            [0x01, 0x01, 0x00, 0x00],  # type
            [0x01, 0x00, 0x00, 0x54],  # version
            [0x39, 0x30, 0x00, 0x00],  # timestamp
            [0x20, 0x00, 0x00, 0x00],  # public key length
            transaction.signer_public_key.bytes,  # signer public key
            [0x60, 0xAE, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00],  # fee
            [0x49, 0x3E, 0x00, 0x00],  # deadline
            [0x28, 0x00, 0x00, 0x00],  # address length
            str(transaction.recipient_address).encode('utf8'),  # recipient address
            [0xFF, 0x5B, 0xB2, 0xEC, 0x22, 0x00, 0x00, 0x00],  # amount
            [0x00, 0x00, 0x00, 0x00]  # message length
        ]
        expected_buffer = reduce(lambda x, y: bytes(x) + bytes(y), expected_buffers)
        self.assertEqual(expected_buffer, buffer)

    def test_can_serialize_with_message(self):
        # Arrange:
        transaction = self._create_transfer_for_serialization_tests(True)

        # Act:
        buffer = transaction.serialize()

        # Assert:
        expected_buffers = [
            [0x01, 0x01, 0x00, 0x00],  # type
            [0x01, 0x00, 0x00, 0x54],  # version
            [0x39, 0x30, 0x00, 0x00],  # timestamp
            [0x20, 0x00, 0x00, 0x00],  # public key length
            transaction.signer_public_key.bytes,  # signer public key
            [0xB0, 0x71, 0x0B, 0x00, 0x00, 0x00, 0x00, 0x00],  # fee
            [0x49, 0x3E, 0x00, 0x00],  # deadline
            [0x28, 0x00, 0x00, 0x00],  # address length
            str(transaction.recipient_address).encode('utf8'),  # recipient address
            [0xFF, 0x5B, 0xB2, 0xEC, 0x22, 0x00, 0x00, 0x00],  # amount
            [0x0F, 0x00, 0x00, 0x00],  # message length (including message header)
            [0x01, 0x00, 0x00, 0x00],  # message type
            [0x07, 0x00, 0x00, 0x00],  # message length
            transaction.message
        ]
        expected_buffer = reduce(lambda x, y: bytes(x) + bytes(y), expected_buffers)
        self.assertEqual(expected_buffer, buffer)

    # endregion

    # region string

    def test_can_create_string_representation_without_message(self):
        # Arrange:
        transaction = self._create_transfer_for_serialization_tests(False)

        # Act:
        transaction_str = str(transaction)

        # Assert:
        expected_transaction_str = '\n'.join([
            '             type = 257 [0x101]',
            '          version = 1409286145 [0x54000001]',
            '        timestamp = 12345 [0x3039]',
            'signer_public_key = D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0',
            '              fee = 700000 [0xAAE60]',
            '         deadline = 15945 [0x3E49]',
            'recipient_address = TCFGSLITSWMRROU2GO7FPMIUUDELUPSZUNUEZF33',
            '           amount = 149999999999 [0x22ECB25BFF]',
            '          message = None'
        ])
        self.assertEqual(expected_transaction_str, transaction_str)

    def test_can_create_string_representation_with_message(self):
        # Arrange:
        transaction = self._create_transfer_for_serialization_tests(True)

        # Act:
        transaction_str = str(transaction)

        # Assert:
        expected_transaction_str = '\n'.join([
            '             type = 257 [0x101]',
            '          version = 1409286145 [0x54000001]',
            '        timestamp = 12345 [0x3039]',
            'signer_public_key = D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0',
            '              fee = 750000 [0xB71B0]',
            '         deadline = 15945 [0x3E49]',
            'recipient_address = TCFGSLITSWMRROU2GO7FPMIUUDELUPSZUNUEZF33',
            '           amount = 149999999999 [0x22ECB25BFF]',
            '          message = 4455981271AB72'
        ])
        self.assertEqual(expected_transaction_str, transaction_str)

    def test_can_create_string_representation_with_message_string(self):
        # Arrange:
        transaction = self._create_transfer_for_serialization_tests(True)
        transaction.message = 'Hello!!'

        # Act:
        transaction_str = str(transaction)

        # Assert:
        expected_transaction_str = '\n'.join([
            '             type = 257 [0x101]',
            '          version = 1409286145 [0x54000001]',
            '        timestamp = 12345 [0x3039]',
            'signer_public_key = D6C3845431236C5A5A907A9E45BD60DA0E12EFD350B970E7F58E3499E2E7A2F0',
            '              fee = 750000 [0xB71B0]',
            '         deadline = 15945 [0x3E49]',
            'recipient_address = TCFGSLITSWMRROU2GO7FPMIUUDELUPSZUNUEZF33',
            '           amount = 149999999999 [0x22ECB25BFF]',
            '          message = Hello!!'
        ])
        self.assertEqual(expected_transaction_str, transaction_str)

    # endregion



import unittest

from symbolchain.core.CryptoTypes import Hash256
from symbolchain.core.sym.MerkleHashBuilder import MerkleHashBuilder
from symbolchain.tests.test.NemTestUtils import NemTestUtils


class MerkleHashBuilderTest(unittest.TestCase):
    def test_can_build_from_zero_hashes(self):
        # Act:
        merkle_hash = self._calculate_merkle_hash([])

        # Assert:
        self.assertEqual(Hash256.zero(), merkle_hash)

    def test_can_build_from_one_hash(self):
        # Arrange:
        seed_hash = NemTestUtils.randcryptotype(Hash256)

        # Act:
        merkle_hash = self._calculate_merkle_hash([seed_hash])

        # Assert:
        self.assertEqual(seed_hash, merkle_hash)

    def test_can_build_from_balanced_tree(self):
        # Act:
        merkle_hash = self._calculate_merkle_hash([
            Hash256('36C8213162CDBC78767CF43D4E06DDBE0D3367B6CEAEAEB577A50E2052441BC8'),
            Hash256('8A316E48F35CDADD3F827663F7535E840289A16A43E7134B053A86773E474C28'),
            Hash256('6D80E71F00DFB73B358B772AD453AEB652AE347D3E098AE269005A88DA0B84A7'),
            Hash256('2AE2CA59B5BB29721BFB79FE113929B6E52891CAA29CBF562EBEDC46903FF681'),
            Hash256('421D6B68A6DF8BB1D5C9ACF7ED44515E77945D42A491BECE68DA009B551EE6CE'),
            Hash256('7A1711AF5C402CFEFF87F6DA4B9C738100A7AC3EDAD38D698DF36CA3FE883480'),
            Hash256('1E6516B2CC617E919FAE0CF8472BEB2BFF598F19C7A7A7DC260BC6715382822C'),
            Hash256('410330530D04A277A7C96C1E4F34184FDEB0FFDA63563EFD796C404D7A6E5A20'),
        ])

        # Assert:
        self.assertEqual(Hash256('7D853079F5F9EE30BDAE49C4956AF20CDF989647AFE971C069AC263DA1FFDF7E'), merkle_hash)

    def test_can_build_from_unbalanced_tree(self):
        # Act:
        merkle_hash = self._calculate_merkle_hash([
            Hash256('36C8213162CDBC78767CF43D4E06DDBE0D3367B6CEAEAEB577A50E2052441BC8'),
            Hash256('8A316E48F35CDADD3F827663F7535E840289A16A43E7134B053A86773E474C28'),
            Hash256('6D80E71F00DFB73B358B772AD453AEB652AE347D3E098AE269005A88DA0B84A7'),
            Hash256('2AE2CA59B5BB29721BFB79FE113929B6E52891CAA29CBF562EBEDC46903FF681'),
            Hash256('421D6B68A6DF8BB1D5C9ACF7ED44515E77945D42A491BECE68DA009B551EE6CE')
        ])

        # Assert:
        self.assertEqual(Hash256('DEFB4BF7ACF2145500087A02C88F8D1FCF27B8DEF4E0FDABE09413D87A3F0D09'), merkle_hash)

    def test_changing_sub_hash_order_changes_merkle_hash(self):
        # Arrange:
        seed_hashes1 = [NemTestUtils.randcryptotype(Hash256) for _ in range(0, 8)]
        seed_hashes2 = [seed_hashes1[i] for i in [0, 1, 2, 5, 4, 3, 6, 7]]

        # Sanity:
        self.assertEqual(len(seed_hashes1), len(seed_hashes2))

        # Act:
        merkle_hash1 = self._calculate_merkle_hash(seed_hashes1)
        merkle_hash2 = self._calculate_merkle_hash(seed_hashes2)

        # Assert:
        self.assertNotEqual(merkle_hash1, merkle_hash2)

    def test_changing_sub_hash_changes_merkle_hash(self):
        # Arrange:
        seed_hashes1 = [NemTestUtils.randcryptotype(Hash256) for _ in range(0, 8)]
        seed_hashes2 = [seed_hashes1[i] if 0 <= i else NemTestUtils.randcryptotype(Hash256) for i in [0, 1, 2, 3, -1, 5, 6, 7]]

        # Sanity:
        self.assertEqual(len(seed_hashes1), len(seed_hashes2))

        # Act:
        merkle_hash1 = self._calculate_merkle_hash(seed_hashes1)
        merkle_hash2 = self._calculate_merkle_hash(seed_hashes2)

        # Assert:
        self.assertNotEqual(merkle_hash1, merkle_hash2)

    @staticmethod
    def _calculate_merkle_hash(seed_hashes):
        builder = MerkleHashBuilder()

        for seed_hash in seed_hashes:
            builder.update(seed_hash)

        return builder.final()



import unittest

from symbolchain.core.BlockchainSettings import BlockchainSettings
from symbolchain.core.CryptoTypes import PublicKey

PUBLIC_KEY_1 = PublicKey('A59277D56E9F4FA46854F5EFAAA253B09F8AE69A473565E01FD9E6A738E4AB74')
PUBLIC_KEY_2 = PublicKey('9A755901AA014A4EACAE615523D2B50C27F954CB936927331F1116C8D5B7B2AA')
YAML_INPUT = '''
blockchain: nis1
network: testnet

nodes:
  - host: alice
    roles: [historical, XXL]

  - host: bob
    roles: [backup, super]

  - host: charlie

accounts:
  - public_key: {public_key_1}
    name: bobby
    roles: [green, main]

  - public_key: {public_key_2}
    name: TEST1
    roles: [red, test]
'''.format(public_key_1=PUBLIC_KEY_1, public_key_2=PUBLIC_KEY_2)


class BlockchainSettingsTest(unittest.TestCase):
    def test_can_load_settings_file(self):
        # Arrange:
        settings = BlockchainSettings.load_from_yaml(YAML_INPUT)

        # Assert:
        self.assertEqual('nis1', settings.blockchain)
        self.assertEqual('testnet', settings.network)

        self.assertEqual(3, len(settings.nodes.descriptors))
        self.assertEqual(['alice', 'bob', 'charlie'], [descriptor.host for descriptor in settings.nodes.descriptors])

        self.assertEqual(2, len(settings.accounts.descriptors))
        self.assertEqual(PUBLIC_KEY_1, settings.accounts.try_find_by_name('bobby').public_key)
        self.assertEqual(PUBLIC_KEY_2, settings.accounts.try_find_by_name('TEST1').public_key)



from .BufferWriter import BufferWriter
from .CryptoTypes import Hash256, Signature
from .QrStorage import QrStorage


class QrSignatureStorage:
    """Loads and saves signatures as QR codes in a directory."""

    def __init__(self, directory):
        """Creates storage for a directory."""
        self.storage = QrStorage(directory)

    def save(self, name, transaction_hash, signatures):
        """Saves a transaction hash along with attesting signatures."""
        writer = BufferWriter()
        writer.write_bytes(transaction_hash.bytes)
        for signature in signatures:
            writer.write_bytes(signature.bytes)

        self.storage.save(name, writer.buffer)

    def load(self, name):
        """Loads a transaction hash along with attesting signatures."""
        decoded_buffer = self.storage.load(name)

        decoded_buffer_size = len(decoded_buffer)
        if decoded_buffer_size < Hash256.SIZE or 0 != (decoded_buffer_size - Hash256.SIZE) % Signature.SIZE:
            raise ValueError('decoded buffer from QR code has unexpected size {}'.format(decoded_buffer_size))

        transaction_hash = Hash256(decoded_buffer[0:Hash256.SIZE])

        signatures = []
        for i in range(0, (decoded_buffer_size - Hash256.SIZE) // Signature.SIZE):
            signature_start = Hash256.SIZE + i * Signature.SIZE
            signatures.append(Signature(decoded_buffer[signature_start:signature_start + Signature.SIZE]))

        return (transaction_hash, signatures)



import unittest

from symbolchain.core.BufferReader import BufferReader


class BufferReaderTest(unittest.TestCase):
    def test_initially_at_eof_when_buffer_empty(self):
        # Act:
        reader = BufferReader(bytes())

        # Assert:
        self.assertTrue(reader.eof)

    def test_initially_not_at_eof_when_buffer_not_empty(self):
        # Act:
        reader = BufferReader(bytes([0xFF]))

        # Assert:
        self.assertFalse(reader.eof)

    def test_eof_only_when_buffer_fully_read(self):
        # Arrange:
        reader = BufferReader(bytes([0xFF, 0xFF, 0xFF, 0xFF]))

        # Act + Assert: partial read
        reader.read_bytes(3)
        self.assertFalse(reader.eof)

        # Act + Assert: full read
        reader.read_bytes(1)
        self.assertTrue(reader.eof)

    def _assert_can_read_int(self, reader):
        # Act:
        read_int1 = reader.read_int(4)
        read_int2 = reader.read_int(2)
        read_int3 = reader.read_int(2)

        # Assert:
        self.assertTrue(reader.eof)
        self.assertEqual(0x05010203, read_int1)
        self.assertEqual(0x00FF, read_int2)
        self.assertEqual(0xEE00, read_int3)

    def test_can_read_int_endianness_default(self):
        self._assert_can_read_int(BufferReader([0x03, 0x02, 0x01, 0x05, 0xFF, 0x00, 0x00, 0xEE]))

    def test_can_read_int_endianness_little(self):
        self._assert_can_read_int(BufferReader([0x03, 0x02, 0x01, 0x05, 0xFF, 0x00, 0x00, 0xEE], 'little'))

    def test_can_read_int_endianness_big(self):
        self._assert_can_read_int(BufferReader([0x05, 0x01, 0x02, 0x03, 0x00, 0xFF, 0xEE, 0x00], 'big'))

    def test_can_read_string(self):
        # Arrange:
        reader = BufferReader(bytes([0x62, 0x61, 0x72]))

        # Act:
        read_str = reader.read_string(3)

        # Assert:
        self.assertTrue(reader.eof)
        self.assertEqual('bar', read_str)

    def test_can_read_hex_string(self):
        # Arrange:
        reader = BufferReader(bytes([0xC0, 0xAA, 0x00, 0x08]))

        # Act:
        read_hex_str = reader.read_hex_string(4)

        # Assert:
        self.assertTrue(reader.eof)
        self.assertEqual('C0AA0008', read_hex_str)

    def test_can_read_bytes(self):
        # Arrange:
        reader = BufferReader(bytes([0x99, 0x00, 0x40, 0x30, 0x78]))

        # Act:
        read_bytes = reader.read_bytes(5)

        # Assert:
        self.assertTrue(reader.eof)
        self.assertEqual(bytes([0x99, 0x00, 0x40, 0x30, 0x78]), read_bytes)

    def _assert_can_read_all(self, reader):
        # Act:
        read_str = reader.read_string(3)
        read_int = reader.read_int(4)
        read_bytes = reader.read_bytes(5)
        read_hex_string = reader.read_hex_string(4)

        # Assert:
        self.assertTrue(reader.eof)
        self.assertEqual('bar', read_str)
        self.assertEqual(0x05010203, read_int)
        self.assertEqual(bytes([0x99, 0x00, 0x40, 0x30, 0x78]), read_bytes)
        self.assertEqual('C0AA0008', read_hex_string)

    def test_can_read_all_endianness_default(self):
        self._assert_can_read_all(BufferReader(bytes([
            0x62, 0x61, 0x72,
            0x03, 0x02, 0x01, 0x05,
            0x99, 0x00, 0x40, 0x30, 0x78,
            0xC0, 0xAA, 0x00, 0x08
        ])))

    def test_can_read_all_endianness_little(self):
        self._assert_can_read_all(BufferReader(bytes([
            0x62, 0x61, 0x72,
            0x03, 0x02, 0x01, 0x05,
            0x99, 0x00, 0x40, 0x30, 0x78,
            0xC0, 0xAA, 0x00, 0x08
        ]), 'little'))

    def test_can_read_all_endianness_big(self):
        self._assert_can_read_all(BufferReader(bytes([
            0x62, 0x61, 0x72,
            0x05, 0x01, 0x02, 0x03,
            0x99, 0x00, 0x40, 0x30, 0x78,
            0xC0, 0xAA, 0x00, 0x08
        ]), 'big'))



import base64
import unittest

import yaml

from symbolchain.core.AccountDescriptorRepository import AccountDescriptorRepository
from symbolchain.core.ByteArray import ByteArray
from symbolchain.core.CryptoTypes import PublicKey
from symbolchain.tests.test.NemTestUtils import NemTestUtils

PUBLIC_KEY_1 = PublicKey('A59277D56E9F4FA46854F5EFAAA253B09F8AE69A473565E01FD9E6A738E4AB74')
PUBLIC_KEY_2 = PublicKey('9A755901AA014A4EACAE615523D2B50C27F954CB936927331F1116C8D5B7B2AA')
YAML_INPUT = '''
- address: TALIC33PNVKIMNXVOCOQGWLZK52K4XALZBNE2ISF
  name: alice
  roles: [green, main]

- public_key: {public_key_1}
  name: TEST1
  roles: [red, test]

- public_key: {public_key_2}
  name: BOB
  roles: [BLUE, main]

- address: TALICEROONSJCPHC63F52V6FY3SDMSVAEUGHMB7C
  name: charlie
'''.format(public_key_1=PUBLIC_KEY_1, public_key_2=PUBLIC_KEY_2)


# NIS style address to avoid circular import
class MockAddress(ByteArray):
    def __init__(self, address):
        super().__init__(25, base64.b32decode(address), MockAddress)


class AccountDescriptorRepositoryTest(unittest.TestCase):
    # region load and find

    def test_can_load_descriptors_yaml(self):
        # Arrange:
        repository = AccountDescriptorRepository(YAML_INPUT)

        # Assert:
        self.assertEqual(4, len(repository.descriptors))
        self.assertEqual(['alice', 'TEST1', 'BOB', 'charlie'], [descriptor.name for descriptor in repository.descriptors])

    def test_can_load_descriptors_list(self):
        # Arrange:
        repository = AccountDescriptorRepository(yaml.load(YAML_INPUT, Loader=yaml.SafeLoader))

        # Assert:
        self.assertEqual(4, len(repository.descriptors))
        self.assertEqual(['alice', 'TEST1', 'BOB', 'charlie'], [descriptor.name for descriptor in repository.descriptors])

    def test_cannot_find_by_name_when_no_match(self):
        # Arrange:
        repository = AccountDescriptorRepository(YAML_INPUT)

        # Act:
        descriptor = repository.try_find_by_name('dave')

        # Assert:
        self.assertEqual(None, descriptor)

    def test_can_find_by_name_when_match(self):
        # Arrange:
        repository = AccountDescriptorRepository(YAML_INPUT)

        # Act
        descriptor1 = repository.try_find_by_name('alice')
        descriptor2 = repository.try_find_by_name('TEST1')
        descriptor3 = repository.try_find_by_name('BOB')
        descriptor4 = repository.try_find_by_name('charlie')

        # Assert:
        self.assertEqual(None, descriptor1.public_key)
        self.assertEqual('TALIC33PNVKIMNXVOCOQGWLZK52K4XALZBNE2ISF', descriptor1.address)
        self.assertEqual('alice', descriptor1.name)
        self.assertEqual(['green', 'main'], descriptor1.roles)

        self.assertEqual(PUBLIC_KEY_1, descriptor2.public_key)
        self.assertEqual(None, descriptor2.address)
        self.assertEqual('TEST1', descriptor2.name)
        self.assertEqual(['red', 'test'], descriptor2.roles)

        self.assertEqual(PUBLIC_KEY_2, descriptor3.public_key)
        self.assertEqual(None, descriptor3.address)
        self.assertEqual('BOB', descriptor3.name)
        self.assertEqual(['BLUE', 'main'], descriptor3.roles)

        self.assertEqual(None, descriptor4.public_key)
        self.assertEqual('TALICEROONSJCPHC63F52V6FY3SDMSVAEUGHMB7C', descriptor4.address)
        self.assertEqual('charlie', descriptor4.name)
        self.assertEqual([], descriptor4.roles)

    def test_cannot_find_by_public_key_when_no_match(self):
        # Arrange:
        repository = AccountDescriptorRepository(YAML_INPUT)

        # Act + Assert:
        with self.assertRaises(StopIteration):
            repository.find_by_public_key(NemTestUtils.randcryptotype(PublicKey))

    def test_can_find_by_public_key_when_match(self):
        # Arrange:
        repository = AccountDescriptorRepository(YAML_INPUT)

        # Act:
        descriptor1 = repository.find_by_public_key(PUBLIC_KEY_1)
        descriptor2 = repository.find_by_public_key(PUBLIC_KEY_2)

        # Assert:
        self.assertEqual('TEST1', descriptor1.name)
        self.assertEqual('BOB', descriptor2.name)

    def _assert_can_find_all_by_role(self, role, expected_match_names):
        # Arrange:
        repository = AccountDescriptorRepository(YAML_INPUT)

        # Act:
        descriptors = repository.find_all_by_role(role)
        match_names = [descriptor.name for descriptor in descriptors]

        # Assert:
        self.assertEqual(expected_match_names, match_names)

    def test_find_all_by_role_when_no_match(self):
        self._assert_can_find_all_by_role('blue', [])

    def test_find_all_by_role_when_single_match(self):
        self._assert_can_find_all_by_role('test', ['TEST1'])

    def test_find_all_by_role_when_multiple_matches(self):
        self._assert_can_find_all_by_role('main', ['alice', 'BOB'])

    def test_find_all_by_role_when_no_role_filter_is_provided(self):
        self._assert_can_find_all_by_role(None, ['alice', 'TEST1', 'BOB', 'charlie'])

    # endregion

    # region to_type_parsing_rules_map

    @staticmethod
    def _create_type_parsing_rules():
        return AccountDescriptorRepository(YAML_INPUT).to_type_parsing_rules_map({
            MockAddress: 'address',
            PublicKey: 'public_key'
        })

    def _assert_can_parse_address(self, value, expected_value):
        # Arrange:
        type_parsing_rules = self._create_type_parsing_rules()

        # Act:
        value = type_parsing_rules[MockAddress](value)

        # Assert:
        self.assertEqual(expected_value, value)

    def test_can_parse_named_address_with_overrride(self):
        self._assert_can_parse_address('alice', MockAddress('TALIC33PNVKIMNXVOCOQGWLZK52K4XALZBNE2ISF'))

    def test_cannot_parse_named_address_without_overrride(self):
        # Arrange:
        type_parsing_rules = self._create_type_parsing_rules()

        # Act + Assert:
        with self.assertRaises(TypeError):
            type_parsing_rules[MockAddress]('TEST1')

    def test_can_parse_unnamed_address(self):
        self._assert_can_parse_address('TAVOZX4HDVOAR4W6K4WJHWPD3MOFU27DFEJDR2PR', MockAddress('TAVOZX4HDVOAR4W6K4WJHWPD3MOFU27DFEJDR2PR'))

    def _assert_can_parse_public_key(self, value, expected_value):
        # Arrange:
        type_parsing_rules = self._create_type_parsing_rules()

        # Act:
        value = type_parsing_rules[PublicKey](value)

        # Assert:
        self.assertEqual(expected_value, value)

    def test_can_parse_named_public_key_with_overrride(self):
        self._assert_can_parse_public_key('TEST1', PUBLIC_KEY_1)

    def test_cannot_parse_named_public_key_without_overrride(self):
        # Arrange:
        type_parsing_rules = self._create_type_parsing_rules()

        # Act + Assert:
        with self.assertRaises(TypeError):
            type_parsing_rules[PublicKey]('alice')

    def test_can_parse_unnamed_public_key(self):
        self._assert_can_parse_public_key(str(PUBLIC_KEY_2), PUBLIC_KEY_2)

    # endregion



import datetime
import unittest

from symbolchain.core.nis1.NetworkTimestamp import NetworkTimestamp
from symbolchain.tests.test.BasicNetworkTimestampTest import BasicNetworkTimestampTest, NetworkTimestampTestDescriptor


class NetworkTimestampTest(BasicNetworkTimestampTest, unittest.TestCase):
    @staticmethod
    def get_test_descriptor():
        epoch_time = datetime.datetime(2015, 3, 29, 0, 6, 25, tzinfo=datetime.timezone.utc)
        return NetworkTimestampTestDescriptor(NetworkTimestamp, epoch_time, 'seconds')

    def test_can_add_seconds(self):
        # Arrange:
        timestamp = NetworkTimestamp(100)

        # Act:
        new_timestamp = timestamp.add_seconds(50)

        # Assert:
        self.assertEqual(100, timestamp.timestamp)
        self.assertEqual(100 + 50, new_timestamp.timestamp)



import unittest

import yaml

from symbolchain.core.NodeDescriptorRepository import NodeDescriptorRepository

YAML_INPUT = '''
- host: ALICE
  roles: [historical, XXL]

- host: bob
  roles: [backup, super]

- host: charlie

- host: mercury
  roles: [super]
'''


class NodeDescriptorRepositoryTest(unittest.TestCase):
    def test_can_load_descriptors_yaml(self):
        # Arrange:
        repository = NodeDescriptorRepository(YAML_INPUT)

        # Assert:
        self.assertEqual(4, len(repository.descriptors))
        self.assertEqual(['ALICE', 'bob', 'charlie', 'mercury'], [descriptor.host for descriptor in repository.descriptors])

    def test_can_load_descriptors_list(self):
        # Arrange:
        repository = NodeDescriptorRepository(yaml.load(YAML_INPUT, Loader=yaml.SafeLoader))

        # Assert:
        self.assertEqual(4, len(repository.descriptors))
        self.assertEqual(['ALICE', 'bob', 'charlie', 'mercury'], [descriptor.host for descriptor in repository.descriptors])

    def _assert_can_find_all_by_role(self, role, expected_match_hosts):
        # Arrange:
        repository = NodeDescriptorRepository(YAML_INPUT)

        # Act:
        descriptors = repository.find_all_by_role(role)
        match_hosts = [descriptor.host for descriptor in descriptors]

        # Assert:
        self.assertEqual(expected_match_hosts, match_hosts)

    def test_find_all_by_role_when_no_match(self):
        self._assert_can_find_all_by_role('regular', [])

    def test_find_all_by_role_when_single_match(self):
        self._assert_can_find_all_by_role('historical', ['ALICE'])

    def test_find_all_by_role_when_multiple_matches(self):
        self._assert_can_find_all_by_role('super', ['bob', 'mercury'])

    def test_find_all_by_role_when_no_role_filter_is_provided(self):
        self._assert_can_find_all_by_role(None, ['ALICE', 'bob', 'charlie', 'mercury'])

    def _assert_can_find_all_not_by_role(self, role, expected_match_hosts):
        # Arrange:
        repository = NodeDescriptorRepository(YAML_INPUT)

        # Act:
        descriptors = repository.find_all_not_by_role(role)
        match_hosts = [descriptor.host for descriptor in descriptors]

        # Assert:
        self.assertEqual(expected_match_hosts, match_hosts)

    def test_find_all_not_by_role_when_no_match(self):
        self._assert_can_find_all_not_by_role('regular', ['ALICE', 'bob', 'charlie', 'mercury'])

    def test_find_all_not_by_role_when_single_match(self):
        self._assert_can_find_all_not_by_role('historical', ['bob', 'charlie', 'mercury'])

    def test_find_all_not_by_role_when_multiple_matches(self):
        self._assert_can_find_all_not_by_role('super', ['ALICE', 'charlie'])

    def test_find_all_not_by_role_when_no_role_filter_is_provided(self):
        self._assert_can_find_all_not_by_role(None, [])



import datetime
from abc import abstractmethod

CUSTOM_EPOCH_TIME = datetime.datetime(2021, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)


class NetworkTimestampTestDescriptor:
    def __init__(self, network_timestamp_class, epoch, time_units):
        self.network_timestamp_class = network_timestamp_class
        self.epoch = epoch
        self.time_units = time_units


class BasicNetworkTimestampTest:
    # pylint: disable=no-member

    # region to_datetime

    def test_can_convert_epochal_timestamp_to_datetime(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        utc_timestamp = test_descriptor.network_timestamp_class(0).to_datetime()

        # Assert:
        self.assertEqual(test_descriptor.epoch, utc_timestamp)

    def test_can_convert_non_epochal_timestamp_to_datetime(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        utc_timestamp = test_descriptor.network_timestamp_class(123).to_datetime()

        # Assert:
        self.assertEqual(test_descriptor.epoch + self._get_time_delta(123), utc_timestamp)

    def test_can_convert_epochal_timestamp_to_datetime_custom_epoch(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        utc_timestamp = test_descriptor.network_timestamp_class(0).to_datetime(CUSTOM_EPOCH_TIME)

        # Assert:
        self.assertEqual(CUSTOM_EPOCH_TIME, utc_timestamp)

    def test_can_convert_non_epochal_timestamp_to_datetime_custom_epoch(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        utc_timestamp = test_descriptor.network_timestamp_class(123).to_datetime(CUSTOM_EPOCH_TIME)

        # Assert:
        self.assertEqual(CUSTOM_EPOCH_TIME + self._get_time_delta(123), utc_timestamp)

    # endregion

    # region from_datetime

    def test_can_convert_datetime_to_epochal_timestamp(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        timestamp = test_descriptor.network_timestamp_class.from_datetime(test_descriptor.epoch)

        # Assert:
        self.assertTrue(timestamp.epochal)
        self.assertEqual(0, timestamp.timestamp)

    def test_can_convert_datetime_to_non_epochal_timestamp(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        timestamp = test_descriptor.network_timestamp_class.from_datetime(test_descriptor.epoch + self._get_time_delta(123))

        # Assert:
        self.assertFalse(timestamp.epochal)
        self.assertEqual(123, timestamp.timestamp)

    def test_can_convert_datetime_to_epochal_timestamp_custom_epoch(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        timestamp = test_descriptor.network_timestamp_class.from_datetime(CUSTOM_EPOCH_TIME, CUSTOM_EPOCH_TIME)

        # Assert:
        self.assertTrue(timestamp.epochal)
        self.assertEqual(0, timestamp.timestamp)

    def test_can_convert_datetime_to_non_epochal_timestamp_custom_epoch(self):
        # Arrange:
        test_descriptor = self.get_test_descriptor()

        # Act:
        timestamp = test_descriptor.network_timestamp_class.from_datetime(CUSTOM_EPOCH_TIME + self._get_time_delta(123), CUSTOM_EPOCH_TIME)

        # Assert:
        self.assertFalse(timestamp.epochal)
        self.assertEqual(123, timestamp.timestamp)

    # endregion

    def _get_time_delta(self, count):
        return datetime.timedelta(**{self.get_test_descriptor().time_units: count})

    @abstractmethod
    def get_test_descriptor(self):
        pass



from setuptools import find_packages, setup

with open('README.md', 'r') as readme_file:
    README = readme_file.read()

NAME = 'symbol-sdk-core-python'

with open('version.txt', 'r') as version_file:
    VERSION = version_file.read().strip()

REQUIRES = [
    'catbuffer==1.0.0',
    'cryptography==3.4.6',
    'mnemonic==0.19',
    'Pillow==8.1.1',
    'pysha3==1.0.2',
    'PyYAML==5.4.1',
    'pyzbar==0.1.8',
    'qrcode==6.1'
]

setup(
    name=NAME,
    version=VERSION,
    description='Symbol SDK Core',
    author='NEM Group',
    author_email='dev@nem.software',
    url='https://github.com/nemtech/symbol-sdk-core-python',
    keywords=['symbol', 'sdk', 'sdk-core', 'Symbol SDK core'],
    install_requires=REQUIRES,
    package_dir={'': '.'},
    packages=find_packages('.'),
    include_package_data=True,
    license='MIT',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers=[
        'Programming Language :: Python :: 3.7',
    ]
)



import base64
import os

import qrcode
from PIL import Image
from pyzbar import pyzbar


class QrStorage:
    """Loads and saves binary data as QR codes in a directory."""

    def __init__(self, directory):
        """Creates storage for a directory."""
        self.directory = directory

    def save(self, name, buffer):
        """Saves a buffer."""
        qrcode_payload = self._binary_to_qrcode_payload(buffer)
        qrcode_builder = qrcode.QRCode(error_correction=qrcode.constants.ERROR_CORRECT_H)
        qrcode_builder.add_data(qrcode.util.QRData(qrcode_payload, qrcode.util.MODE_ALPHA_NUM))
        qrcode_builder.make(fit=True)

        qrcode_image = qrcode_builder.make_image(fill_color='#44004E', back_color='#FFFFFF')
        qrcode_image.save(self._get_file_path(name))

    def load(self, name):
        """Loads a buffer."""
        # [0] because exactly one qrcode is expected in the image
        qrcode_payload = pyzbar.decode(Image.open(self._get_file_path(name)), symbols=[pyzbar.ZBarSymbol.QRCODE])[0].data
        return self._qrcode_payload_to_binary(qrcode_payload)

    @staticmethod
    def _binary_to_qrcode_payload(buffer):
        # replace '=' with a character in QR code alphanumeric alphabet, which enables more efficient data packing in the qrcode
        return base64.b32encode(buffer).decode('utf8').replace('=', '$').encode('utf8')

    @staticmethod
    def _qrcode_payload_to_binary(qrcode_payload):
        return base64.b32decode(qrcode_payload.decode('utf8').replace('$', '='))

    def _get_file_path(self, name):
        return os.path.join(self.directory, '{}.png'.format(name))



import unittest

from symbolchain.core.CryptoTypes import Hash256, PrivateKey, PublicKey, Signature
from symbolchain.tests.test.NemTestUtils import NemTestUtils


class CryptoTypesTest(unittest.TestCase):
    # Hash256

    def test_can_create_hash256_with_correct_number_of_bytes(self):
        self._assert_can_create_byte_array_with_correct_number_of_bytes(Hash256, 32)

    def test_cannot_create_hash256_with_incorrect_number_of_bytes(self):
        self._assert_cannot_create_byte_array_with_incorrect_number_of_bytes(Hash256, 32)

    def test_can_create_zeroed_hash256(self):
        self._assert_can_create_zeroed_byte_array(Hash256, 32)

    # endregion

    # PrivateKey

    def test_can_create_private_key_with_correct_number_of_bytes(self):
        self._assert_can_create_byte_array_with_correct_number_of_bytes(PrivateKey, 32)

    def test_cannot_create_private_key_with_incorrect_number_of_bytes(self):
        self._assert_cannot_create_byte_array_with_incorrect_number_of_bytes(PrivateKey, 32)

    def test_can_create_random_private_key(self):
        # Act:
        private_key1 = PrivateKey.random()
        private_key2 = PrivateKey.random()

        # Assert:
        self.assertNotEqual(private_key1, private_key2)

    # endregion

    # region PublicKey

    def test_can_create_public_key_with_correct_number_of_bytes(self):
        self._assert_can_create_byte_array_with_correct_number_of_bytes(PublicKey, 32)

    def test_cannot_create_public_key_with_incorrect_number_of_bytes(self):
        self._assert_cannot_create_byte_array_with_incorrect_number_of_bytes(PublicKey, 32)

    def test_can_create_public_key_from_existing_public_key(self):
        # Arrange:
        raw_bytes = NemTestUtils.randbytes(32)

        # Act:
        byte_array = PublicKey(PublicKey(raw_bytes))

        # Assert:
        self.assertEqual(raw_bytes, byte_array.bytes)

    # endregion

    # region Signature

    def test_can_create_signature_with_correct_number_of_bytes(self):
        self._assert_can_create_byte_array_with_correct_number_of_bytes(Signature, 64)

    def test_cannot_create_signature_with_incorrect_number_of_bytes(self):
        self._assert_cannot_create_byte_array_with_incorrect_number_of_bytes(Signature, 64)

    def test_can_create_zeroed_signature(self):
        self._assert_can_create_zeroed_byte_array(Signature, 64)

    # endregion

    def test_equality_is_only_possible_for_same_types(self):
        # Arrange:
        raw_bytes = NemTestUtils.randbytes(32)

        # Act + Assert:
        self.assertEqual(Hash256(raw_bytes), Hash256(raw_bytes))
        self.assertNotEqual(Hash256(raw_bytes), PrivateKey(raw_bytes))
        self.assertNotEqual(Hash256(raw_bytes), PublicKey(raw_bytes))

        self.assertNotEqual(PrivateKey(raw_bytes), Hash256(raw_bytes))
        self.assertEqual(PrivateKey(raw_bytes), PrivateKey(raw_bytes))
        self.assertNotEqual(PrivateKey(raw_bytes), PublicKey(raw_bytes))

        self.assertNotEqual(PublicKey(raw_bytes), Hash256(raw_bytes))
        self.assertNotEqual(PublicKey(raw_bytes), PrivateKey(raw_bytes))
        self.assertEqual(PublicKey(raw_bytes), PublicKey(raw_bytes))

    def _assert_can_create_byte_array_with_correct_number_of_bytes(self, byte_array_class, size):
        # Arrange:
        raw_bytes = NemTestUtils.randbytes(size)

        # Act:
        byte_array = byte_array_class(raw_bytes)

        # Assert:
        self.assertEqual(raw_bytes, byte_array.bytes)

    def _assert_cannot_create_byte_array_with_incorrect_number_of_bytes(self, byte_array_class, required_size):
        for size in [0, required_size - 1, required_size + 1]:
            with self.assertRaises(ValueError):
                byte_array_class(NemTestUtils.randbytes(size))

    def _assert_can_create_zeroed_byte_array(self, byte_array_class, size):
        # Act:
        byte_array = byte_array_class.zero()

        # Assert:
        self.assertEqual(byte_array_class(bytes([0] * size)), byte_array)



from ..BufferWriter import BufferWriter
from ..CryptoTypes import PrivateKey
from .KeyPair import KeyPair


class VotingKeysGenerator:
    """Generates symbol voting keys."""

    def __init__(self, root_key_pair, private_key_generator=PrivateKey.random):
        """Creates a generator around a voting root key pair."""
        self.root_key_pair = root_key_pair
        self.private_key_generator = private_key_generator

    def generate(self, start_epoch, end_epoch):
        """Generates voting keys for specified epochs."""
        writer = BufferWriter()
        writer.write_int(start_epoch, 8)
        writer.write_int(end_epoch, 8)
        writer.write_int(0xFFFFFFFFFFFFFFFF, 8)
        writer.write_int(0xFFFFFFFFFFFFFFFF, 8)

        writer.write_bytes(self.root_key_pair.public_key.bytes)
        writer.write_int(start_epoch, 8)
        writer.write_int(end_epoch, 8)

        for identifier in reversed(range(start_epoch, end_epoch + 1)):
            child_private_key = self.private_key_generator()
            child_key_pair = KeyPair(child_private_key)

            parent_signed_payload_writer = BufferWriter()
            parent_signed_payload_writer.write_bytes(child_key_pair.public_key.bytes)
            parent_signed_payload_writer.write_int(identifier, 8)
            signature = self.root_key_pair.sign(parent_signed_payload_writer.buffer)

            writer.write_bytes(child_key_pair.private_key.bytes)
            writer.write_bytes(signature.bytes)

        return writer.buffer



import unittest
from binascii import unhexlify

from symbolchain.core.AccountDescriptorRepository import AccountDescriptorRepository
from symbolchain.core.Bip32 import Bip32
from symbolchain.core.CryptoTypes import Hash256, PrivateKey, PublicKey, Signature
from symbolchain.core.facade.SymFacade import SymFacade
from symbolchain.tests.test.NemTestUtils import NemTestUtils

YAML_INPUT = '''
- public_key: 87DA603E7BE5656C45692D5FC7F6D0EF8F24BB7A5C10ED5FDA8C5CFBC49FCBC8
  name: TEST
'''


class SymFacadeTest(unittest.TestCase):
    # region constants

    def test_bip32_constants_are_correct(self):
        self.assertEqual(4343, SymFacade.BIP32_COIN_ID)
        self.assertEqual('ed25519', SymFacade.BIP32_CURVE_NAME)

    def test_key_pair_is_correct(self):
        # Arrange:
        private_key = PrivateKey('E88283CE35FE74C89FFCB2D8BFA0A2CF6108BDC0D07606DEE34D161C30AC2F1E')

        # Act:
        key_pair = SymFacade.KeyPair(private_key)

        # Assert:
        self.assertEqual(PublicKey('E29C5934F44482E7A9F50725C8681DE6CA63F49E5562DB7E5BC9EABA31356BAD'), key_pair.public_key)

    def test_can_sign_and_verify(self):
        # Arrange:
        private_key = PrivateKey.random()
        key_pair = SymFacade.KeyPair(private_key)
        message = NemTestUtils.randbytes(21)

        # Act:
        signature = key_pair.sign(message)
        is_verified = SymFacade.Verifier(key_pair.public_key).verify(message, signature)

        # Assert:
        self.assertTrue(is_verified)

    # endregion

    # region constructor

    def test_can_create_around_known_network(self):
        # Act:
        facade = SymFacade('public_test')
        transaction = facade.transaction_factory.create({
            'type': 'transfer',
            'signer_public_key': NemTestUtils.randcryptotype(PublicKey)
        })

        # Assert:
        self.assertEqual('public_test', facade.network.name)

        self.assertEqual(0x4154, transaction.type)
        self.assertEqual(1, transaction.version)

    def test_cannot_create_around_unknown_network(self):
        # Act:
        with self.assertRaises(StopIteration):
            SymFacade('foo')

    # endregion

    # region hash_transaction / sign_transaction

    @staticmethod
    def _create_real_transfer(facade):
        return facade.transaction_factory.create({
            'type': 'transfer',
            'signer_public_key': 'TEST',
            'fee': 1000000,
            'deadline': 41998024783,
            'recipient_address': 'TD4PJKW5JP3CNHA47VDFIM25RCWTWRGT45HMPSA',
            'mosaics': [(0x2CF403E85507F39E, 1000000)]
        })

    @staticmethod
    def _create_real_aggregate(facade):
        aggregate = facade.transaction_factory.create({
            'type': 'aggregateComplete',
            'signer_public_key': 'TEST',
            'fee': 2000000,
            'deadline': 42238390163,
            'transactions_hash': unhexlify('71554638F578358B1D3FC4369AC625DB491AD5E5D4424D6DBED9FFC7411A37FE'),
        })
        transfer = facade.transaction_factory.create_embedded({
            'type': 'transfer',
            'signer_public_key': 'TEST',
            'recipient_address': 'TCIDK4CGCHGVZHLNTOKJ32MFEZWMFBCWUJIAXCA',
            'mosaics': [(0x2CF403E85507F39E, 1000000)]
        })
        aggregate.transactions.append(transfer)
        return aggregate

    def test_can_hash_transaction(self):
        # Arrange:
        private_key = PrivateKey('EDB671EB741BD676969D8A035271D1EE5E75DF33278083D877F23615EB839FEC')
        facade = SymFacade('public_test', AccountDescriptorRepository(YAML_INPUT))

        transaction = self._create_real_transfer(facade)
        transaction.signature = facade.sign_transaction(facade.KeyPair(private_key), transaction).bytes

        # Act:
        hash_value = facade.hash_transaction(transaction)

        # Assert:
        self.assertEqual(Hash256('3FAC33913FB3D7CF24618FD654C6635517B3199961062869DF96DA6B5B22F26F'), hash_value)

    def test_can_hash_aggregate_transaction(self):
        # Arrange:
        private_key = PrivateKey('EDB671EB741BD676969D8A035271D1EE5E75DF33278083D877F23615EB839FEC')
        facade = SymFacade('public_test', AccountDescriptorRepository(YAML_INPUT))

        transaction = self._create_real_aggregate(facade)
        transaction.signature = facade.sign_transaction(facade.KeyPair(private_key), transaction).bytes

        # Act:
        hash_value = facade.hash_transaction(transaction)

        # Assert:
        self.assertEqual(Hash256('6B14F18A594FEEEE14731FD381C45991D9A59B7BC948EAE9DF2D977ABEAF0FBE'), hash_value)

    def test_can_sign_transaction(self):
        # Arrange:
        private_key = PrivateKey('EDB671EB741BD676969D8A035271D1EE5E75DF33278083D877F23615EB839FEC')
        facade = SymFacade('public_test', AccountDescriptorRepository(YAML_INPUT))

        transaction = self._create_real_transfer(facade)

        # Sanity:
        self.assertEqual(Signature.zero().bytes, transaction.signature)

        # Act:
        signature = facade.sign_transaction(facade.KeyPair(private_key), transaction)

        # Assert:
        expected_signature = Signature(''.join([
            'FD112693296E28EED2438C75DFDDDFBC87B8F9A51AEAF7806014D93AE89893BD',
            'AD6E74DA06C3502632A6C905E674DA6E13B8C6DFCB198504546EAD01294AB104'
        ]))
        self.assertEqual(expected_signature, signature)

    def test_can_sign_aggregate_transaction(self):
        # Arrange:
        private_key = PrivateKey('EDB671EB741BD676969D8A035271D1EE5E75DF33278083D877F23615EB839FEC')
        facade = SymFacade('public_test', AccountDescriptorRepository(YAML_INPUT))

        transaction = self._create_real_aggregate(facade)

        # Sanity:
        self.assertEqual(Signature.zero().bytes, transaction.signature)

        # Act:
        signature = facade.sign_transaction(facade.KeyPair(private_key), transaction)

        # Assert:
        expected_signature = Signature(''.join([
            '76597917A1CFB587A74329432CE84460F73E36E1C4DBB968E5849116BE87B8CC',
            '96B273F2C0506CDE44CDD924A9A5802019F140B56EDDB0AF73277651C9DA8008'
        ]))
        self.assertEqual(expected_signature, signature)

    def _assert_can_verify_transaction(self, transaction_factory):
        # Arrange:
        private_key = PrivateKey('EDB671EB741BD676969D8A035271D1EE5E75DF33278083D877F23615EB839FEC')
        facade = SymFacade('public_test', AccountDescriptorRepository(YAML_INPUT))

        transaction = transaction_factory(facade)

        # Sanity:
        self.assertEqual(Signature.zero().bytes, transaction.signature)

        # Act:
        signature = facade.sign_transaction(facade.KeyPair(private_key), transaction)
        is_verified = facade.verify_transaction(transaction, signature)

        # Assert:
        self.assertTrue(is_verified)

    def test_can_verify_transaction(self):
        self._assert_can_verify_transaction(self._create_real_transfer)

    def test_can_verify_aggregate_transaction(self):
        self._assert_can_verify_transaction(self._create_real_aggregate)

    # endregion

    # region bip32_node_to_key_pair

    def _assert_bip32_child_public_keys(self, passphrase, expected_child_public_keys):
        # Arrange:
        mnemonic_seed = ' '.join([
            'hamster', 'diagram', 'private', 'dutch', 'cause', 'delay', 'private', 'meat', 'slide', 'toddler', 'razor', 'book',
            'happy', 'fancy', 'gospel', 'tennis', 'maple', 'dilemma', 'loan', 'word', 'shrug', 'inflict', 'delay', 'length'
        ])

        # Act:
        root_node = Bip32(SymFacade.BIP32_CURVE_NAME).from_mnemonic(mnemonic_seed, passphrase)

        child_public_keys = []
        for i in range(0, len(expected_child_public_keys)):
            child_node = root_node.derive_path([44, SymFacade.BIP32_COIN_ID, i, 0, 0])
            child_key_pair = SymFacade.bip32_node_to_key_pair(child_node)
            child_public_keys.append(child_key_pair.public_key)

        # Assert:
        self.assertEqual(expected_child_public_keys, child_public_keys)

    def test_can_use_bip32_derivation_without_passphrase(self):
        self._assert_bip32_child_public_keys('', [
            PublicKey('E9CFE9F59CB4393E61B2F42769D9084A644B16883C32C2823E7DF9A3AF83C121'),
            PublicKey('0DE8C3235271E4C9ACF5482F7DFEC1E5C4B20FFC71548703EACF593153F116F9'),
            PublicKey('259866A68A00C325713342232056333D60710E223FC920566B3248B266E899D5')
        ])

    def test_can_use_bip32_derivation_with_passphrase(self):
        self._assert_bip32_child_public_keys('TREZOR', [
            PublicKey('47F4D39D36D11C07735D7BE99220696AAEE7B3EE161D61422220DFE3FF120B15'),
            PublicKey('4BA67E87E8C14F3EB82B3677EA959B56A9D7355705019CED1FCF6C76104E628C'),
            PublicKey('8115D75C13C2D25E7FA3009D03D63F1F32601CDCCA9244D5FDAC74BCF3E892E3')
        ])

    # endregion



from ..CryptoTypes import PublicKey
from .Network import Address
from .Transaction import Transaction

RECIPIENT_LENGTH = 40


class TransferTransaction(Transaction):
    """Represents a transfer transaction."""
    NAME = 'transfer'
    TYPE = 0x0101

    TYPE_HINTS = {
        'signer_public_key': PublicKey,
        'recipient_address': Address
    }

    def __init__(self, network):
        """Creates a transfer transaction for the specified network."""
        super().__init__(network, TransferTransaction.TYPE)

        self.recipient_address = None
        self.amount = 0
        self._message = None

    @property
    def fee(self):
        """Gets the (minimum) fee."""
        fee = min(25, max(1, self.amount // 10000000000))

        if self.message:
            fee += len(self.message) // 32 + 1

        return int(0.05 * fee * 1000000)

    @property
    def message(self):
        """Gets the message."""
        return self._message

    @message.setter
    def message(self, value):
        """Sets the message."""
        if isinstance(value, str):
            self._message = value.encode('utf8')
        else:
            self._message = bytes(value)

    def serialize_custom(self, writer):
        writer.write_int(RECIPIENT_LENGTH, 4)
        writer.write_string(str(self.recipient_address))

        writer.write_int(self.amount, 8)

        if not self.message:
            writer.write_int(0, 4)
        else:
            message_length = len(self.message)

            writer.write_int(message_length + 8, 4)
            writer.write_int(1, 4)
            writer.write_int(message_length, 4)
            writer.write_bytes(self.message)

        return writer.buffer

    @staticmethod
    def field_names():
        return ['recipient_address', 'amount', 'message']
